import 'dart:async';
import 'dart:typed_data';
import 'world.dart';
import 'SitesPreferences.dart';

class ModeratorSites {
  //init site from startup list
  //or join site
  SitesPreferences _preferences = SitesPreferences([]);
  interzoneDisk? disk;
  Map<String, world> _sites = {};
  bool get _dirty {
    for (var s in _sites.values) {
      if (s.isDirty) return true;
    }
    return false;
  }

  void unDirty() {
    for (var s in _sites.values) {
      s.unDirty();
    }
  }

  int get now => DateTime.now().millisecondsSinceEpoch;

  void housekeep() {
    //saves cached thing if needed
    if (_preferences.isDirty) {
      disk?.commitPreferences(_preferences.toJson());
      _preferences.unDirty();
    }
  }

  Set<String> get followedSites {
    Set<String> _l = {};
    _preferences.sites.forEach((element) {
      _l.add(element.siteUrl);
    });
    return _l;
  }

  Future<void> init(interzoneDisk _io) async {
    disk = _io;
    //read app prefs
    await disk!.init('sites', false);
    final prjs = await _io.readPreferences;
    if (prjs.isNotEmpty) {
      _preferences = SitesPreferences.fromJson(prjs);
      if (_preferences.sites.isNotEmpty) {
        //swithc later
        //switchSite(_preferences.sites.first.siteName,
        //    _preferences.sites.first.siteUrl);
      }
    }
    //read site specific things
    final prs = _preferences.recentSite;
    if (prs != null) {
      followSite(prs.siteName, prs.siteUrl, prs.nick, prs.avatar)
          .then((value) async {
        if (value) {
          switchSite(prs.siteName, prs.siteUrl, prs.nick, prs.avatar);
        }
      });
    }
  }

  String _homeSite = '';
  String lastPushedCID = '';
  int lastPushedCIDByteLength = 0;

  List<world> get allSites => _sites.values.toList();
  world? get homeSite => _sites[_homeSite];
  String get homeSiteName =>
      homeSite != null ? homeSite!.siteName : 'NO_HOMESITE';
  Uint8List get xxWants => homeSite != null ? homeSite!.xxWants : Uint8List(0);
  Future<Uint8List> satisfyXXwant(Uint8List data) async =>
      homeSite != null ? await homeSite!.satisfyXXwant(data) : Uint8List(0);
  Future<bool> injectXX(int xx, Uint8List data) async =>
      homeSite != null ? await homeSite!.satisfyXX(xx, data) : false;

  Future<bool> followSite(
      String siteName, String url, String nick, String avatar) async {
    if (!url.startsWith('https://')) {
      url = 'https://' + url.split(' ').first;
    }
    if (siteName.isEmpty) {
      siteName = url.split('/')[2];
    }
    siteName.trim();
    url.trim();
    //if (_preferences.names.contains(siteName)) return true;

    if (_sites[siteName] != null) {
      _sites[siteName]!.refreshViaHTTP();
      return true;
    }
    //follow first, switch later
    await disk!.init(siteName + 'Data', false);
    final world iz = world();
    iz.appSplashImage = siteName;

    //TODO put loginy stuff
    return await iz.init(siteName, url, nick, avatar, disk!).then((value) {
      if (value == null) return false;

      _sites[siteName] = iz;
      _preferences.addSite(siteName, url, nick, avatar, now);

      return true;
    });
  }

  Future<bool> switchSite(String siteName, String url, nick, avatar) async {
    //gets url
    if (siteName.isEmpty) {
      siteName = url.split('/')[2];
    }
    if (!_preferences.names.contains(siteName)) return false;
    final site = _preferences.site(siteName);
    _preferences.setNickAvatar(siteName, nick, avatar, now);

    if (_homeSite == site!.siteName) return true;
    if (_sites[site!.siteName] != null) {
      homeSite?.logout;
      //repull cachoid
      _combinedEntrySetCache.all.clear();
      final lin = _sites[site!.siteName]!.login;
      if (lin != null) {
        _homeSite = siteName;
        
        homeSite?.outgoingMessageCacheSend(0);
        return true;
      }
    }

    return false;

    /*await disk!.init(site!.siteName + 'Data', false);
    final world iz = world();
    iz.appSplashImage = site!.siteName;
    _sites[site!.siteName] = iz;
    _homeSite = site!.siteName;
    if (nick.isNotEmpty) {
      site.nick = nick;
    }
    if (avatar.isNotEmpty) {
      site.avatar = avatar;
    }
    //TODO put loginy stuff
    return await iz
        .init(site!.siteName, site!.siteUrl, site!.nick, site.avatar, disk!)
        .then((value) {
      if (value == null) return false;
      _combinedEntrySetCache = ModeratorEntrySet();
      for (var i in value) {
        _combinedEntrySetCache.add(i);
      }
      _homeSite = site!.siteName;

      /*ModeratorEntrySet _es = ModeratorEntrySet();
      for (world w in _sites.values) {
        for (ModeratorEntry e in w.messages.all) {
          //e.siteUrl = key;
          _es.add(e);
        }
      }
      _combinedEntrySetCache = _es;*/

      return true;
    });

    return false;

    return await followSite(siteName, url, '', 'avatar').then((value) {
      if (value) {
        //homeSite?.logout;
        _homeSite = siteName;
        //repull cachoid
        _combinedEntrySetCache.all.clear();
        return true;
      }
      return false;
    });
    */
  }

  bool ping() {
    allSites.forEach((element) {
      element.ping();
    });
    return true;
  }

  world? findSite(String site) => _sites[site];
  world? get find_homeSite => findSite(_homeSite);
  String get nick {
    if (_preferences.recentSite != null) {
      return _preferences.recentSite!.nick;
    }
    return 'guest';
  }

  String get avatar {
    if (_preferences.recentSite != null) {
      return _preferences.recentSite!.avatar;
    }
    return '🐷';
  }

  ModeratorEntrySet _combinedEntrySetCache = ModeratorEntrySet();

  ModeratorEntrySet get combinedEntrySet {
    if (!_dirty && _combinedEntrySetCache.all.isNotEmpty) {
      return _combinedEntrySetCache;
    }
    ModeratorEntrySet _es = ModeratorEntrySet();
    //_sites.forEach((key, value) {
    homeSite?.messages.all.forEach((element) {
      element.siteUrl = homeSite!.baseUrl;
      _es.add(element);
    });
    //});
    _combinedEntrySetCache = _es;
    unDirty();
    return _combinedEntrySetCache;
  }

  ModeratorEntrySet _latestThreadsEntrySetCache = ModeratorEntrySet();

  ModeratorEntrySet get latestThreadsEntrySet {
    //find stuff with updated stuff possibly
    if (!_dirty && _latestThreadsEntrySetCache.all.isNotEmpty) {
      return _latestThreadsEntrySetCache;
    }
    ModeratorEntrySet _es = ModeratorEntrySet();

    if (homeSite == null) {
      _latestThreadsEntrySetCache = _es;
      return _es;
    }
    _latestThreadsEntrySetCache = homeSite!.LatestThreads;
    //recalc thread shortcuts
    _tbsCache.clear;
    unDirty();
    return _latestThreadsEntrySetCache;
  }

  ModeratorEntrySet get HomeSiteLatestThreadsEntrySet {
    //find stuff with updated stuff possibly
    if (!_dirty && _latestThreadsEntrySetCache.all.isNotEmpty) {
      return _latestThreadsEntrySetCache;
    }
    ModeratorEntrySet _es = ModeratorEntrySet();

    if (homeSite == null) {
      _latestThreadsEntrySetCache = _es;
      return _es;
    }
    _latestThreadsEntrySetCache = homeSite!.LatestThreads;
    //recalc thread shortcuts
    _tbsCache.clear;
    unDirty();
    return _latestThreadsEntrySetCache;
  }

  ModeratorEntrySet taggedAs(String tag) =>
      ModeratorEntrySet.fromList(latestThreadsEntrySet.taggedAs(tag));
  ModeratorEntrySet taggedAsWithReplies(String shortLink) {
    final t = threadByShortLink(shortLink);
    if (t.isEmpty) return ModeratorEntrySet();
    final tag = t.first.tags.first;

    ModeratorEntrySet _l = ModeratorEntrySet();
    taggedAs(tag).all.forEach((element) {
      _l.add(element);
      threadByShortLink(element.shortLink).forEach((e2) {
        if (e2.flags.isReply) {
          _l.add(e2);
        }
      });
    });
    return _l;
  }

  ModeratorEntrySet AuthoredByWithReplies(String shortLink) {
    final t = threadByShortLink(shortLink);
    if (t.isEmpty) return ModeratorEntrySet();
    final nick = t.first.nick;

    ModeratorEntrySet _l = ModeratorEntrySet();
    authoredBy(nick).all.forEach((element) {
      _l.add(element);
      threadByShortLink(element.shortLink).forEach((e2) {
        if (e2.flags.isReply) {
          _l.add(e2);
        }
      });
    });
    return _l;
  }

  ModeratorEntrySet mentioning(String tag) =>
      ModeratorEntrySet.fromList(latestThreadsEntrySet.mentioning(tag));
  ModeratorEntrySet authoredBy(String tag) =>
      ModeratorEntrySet.fromList(latestThreadsEntrySet.authoredBy(tag));

  Map<String, List<ModeratorEntry>> _tbsCache = {};

  List<ModeratorEntry> threadByShortLink(String shortLink) {
    if (_tbsCache[shortLink] != null) return _tbsCache[shortLink]!;
    Set<ModeratorEntry> _ll = {};

    final s = homeSite?.threadByShortLink(shortLink);
    if (s == null) return _ll.toList();
    if (s != null && s.isNotEmpty) {
      _ll.addAll(s);
    }

    _tbsCache[shortLink] = _ll.toList();
    return _ll.toList();
  }

  List<ModeratorEntry> unreadThreadByShortLink(String shortLink) {
    return latestThreadsEntrySet.threadByShortLink(shortLink).toList();
  }

  ModeratorEntrySet? siteEntrySet(String site) {
    ModeratorEntrySet _es = ModeratorEntrySet();
    if (_sites[site] == null) return _es;
    return _sites[site]?.messages;
  }

  Future<bool> setNickAvatar(String nick, String avatar, String site) async {
    if (_preferences.sites.isNotEmpty) {
      _preferences.setNickAvatar(site, nick, avatar, now);
    }
    await homeSite?.setUserNickAndImage(nick, avatar, Uint8List(0));
    return true;
  }

  Future<bool> setIpfsCredentials(
      String url, String k1, String k2, String site) async {
    if (site.isEmpty) {
      for (var i in _sites.values) {
        i.setIpfsCredentials(k1, k2, url);
      }

      return true;
    }
    return await _sites[site]!.setIpfsCredentials(k1, k2, '');
  }

  Future<bool> unfollowSite(String siteName, String url, String nick) async {
    if (siteName.isEmpty) {
      siteName = url.split('/')[2];
    }
    //todo flush caches etc
    _sites[siteName]!.terminate();
    _sites.remove(siteName);
    if (_preferences.unfollowSite(siteName)) {
      //TODO purge site data
      disk?.commitPreferences(_preferences.toJson());
    }

    if (_homeSite == siteName) {
      _homeSite = _preferences.sites.first.siteName;
    }

    ModeratorEntrySet _es = ModeratorEntrySet();
    _sites.forEach((key, value) {
      value.messages.all.forEach((element) {
        element.siteUrl = key;
        _es.add(element);
      });
    });
    _combinedEntrySetCache = _es;

    return true;
  }

  /*Future<bool> deleteEntry(String shortLink) async {
    if (_sites[_homeSite] == null) return false;
    _sites[_homeSite]!.deleteEntry(shortLink);
    return true;
  }*/

  Future<bool> dropEntry(String shortLink) async {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    if (target == null) return false;
    final surl = target.siteUrl.isNotEmpty ? target.siteUrl : _homeSite;
    final li = _sites[surl]!.dropEntryToQueue(shortLink);
    if (li == null) return false;
    _sites[surl]!.dropEntryOmni(shortLink);
    if (target.flags.isPost) {
      _tbsCache.remove(shortLink);
    }
    if (target.flags.isReply) {
      _tbsCache.clear();
    }
    return true;
  }

  List<int> get droppedAttachments =>
      homeSite != null ? homeSite!.droppedAttachments : [];
  int get attachmentsWeightInBytes =>
      homeSite != null ? homeSite!.attachmentsWeightInBytes : 0;

  bool markSeenOmni(String shortLink) {
    //hack to update cache
    if (_latestThreadsEntrySetCache.markSeenByShortLink(shortLink)) {
      ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
      if (target == null) return false;
      final surl = target.siteUrl.isNotEmpty ? target.siteUrl : _homeSite;
      final li = _sites[surl]!.markSeenOmni(shortLink);
      if (li == null) return false;
    }

    _tbsCache.remove(shortLink);
    return true;
  }

  bool isSeenOmni(String shortLink) {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    if (target == null) return false;
    final surl = target.siteUrl.isNotEmpty ? target.siteUrl : _homeSite;
    return _sites[surl]!.isSeenOmni(shortLink);
  }

  bool peekThreadOmni(String shortLink) {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    if (target == null) return false;
    final surl = target.siteUrl.isNotEmpty ? target.siteUrl : _homeSite;
    final li = _sites[surl]!.peekThreadOmni(shortLink);
    if (li == null) return false;
    return true;
  }

  Future<int> muteAuthor(String shortLink) async {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    if (target == null) return -1;
    final surl = target.siteUrl.isNotEmpty ? target.siteUrl : _homeSite;
    //TODO add OMNIdrops

    return await _sites[surl]!.muteAuthorToQueue(target);
  }

  Future<bool> shareEntry(String shortLink) async {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    if (target == null) return false;
    final surl = target.siteUrl.isNotEmpty ? target.siteUrl : _homeSite;
    final li = await _sites[surl]!.shareEntryToQueue(target);
    if (li == null) return false;
    return true;
  }

  Future<int> likeEntry(String shortLink) async {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    if (target == null) return -1;

    final li = homeSite?.likeEntryToQueue(target);
    if (li == null) return -1;
    return li;
  }

  Future<int> unlikeEntry(String shortLink) async {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    if (target == null) return -1;
    final li = homeSite?.unlikeEntryToQueue(target);
    if (li == null) return -1;
    return li;
  }

  Future<int> postEntry(String shortLink, String text, Uint8List attachment,
      String attachentType) async {
    if (homeSite == null) return -1;

    return await homeSite!.newThreadToQueue(_homeSite, text, attachment, nick,
        avatar, shortLink, '', attachentType);
  }

  Future<int> replyEntry(String shortLink, String replyText, String ppShortLink,
      Uint8List attachement, String attachmentType) async {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    if (target == null) return -1;
    final surl = target.siteUrl.isNotEmpty ? target.siteUrl : _homeSite;

    return _sites[surl]!
        .newReplyToQueue(replyText, target.shortLink, ppShortLink, attachement,
            attachmentType, nick, avatar)
        .then((value) {
      if (value > -1) {
        _tbsCache.remove(shortLink);
      }
      return value;
    });
  }

  Future<List<ModeratorEntry>?> outgoingMessageCacheSend(int timestamp) async {
    final t = await homeSite?.outgoingMessageCacheSend(timestamp);
    if (t != null && t.isNotEmpty) {
      //if we get anything at all, some addition happened?

      _tbsCache.clear();
    }
    return t;
  }

  Future<List<ModeratorEntry>?> refreshThread(String shortLink) async {
    //saves cached thing if needed
    housekeep();

    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    final r = await homeSite?.refreshThreadViaHTTP(shortLink);
    //if offline, get local
    // if (r == null) return threadByShortLink(shortLink);
    if (r != null && r.isNotEmpty) {
      _tbsCache.clear();
    }

    return threadByShortLink(shortLink);
  }

  Future<List<ModeratorEntry>?> refreshGroupThread(String shortLink) async {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);

    final r = await homeSite?.refreshGroupThreadViaHTTP(shortLink);
    //if offline, get local
    if (r == null) return threadByShortLink(shortLink);
    _tbsCache.clear();
    return r;
  }

  Future<List<ModeratorEntry>?> refreshTagged(String shortLink) async {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    //if thread missing try to find on serv
    //if (target == null) return false;
    //final surl = target?.siteUrl.isNotEmpty ? target.siteUrl : _homeSite;
    final surl = _homeSite;
    final r = await homeSite?.refreshTaggedViaHTTP(shortLink);
    if (r == null) return r;
    _tbsCache.clear();
    return r;
  }

  Future<List<ModeratorEntry>?> refreshAuthor(String shortLink) async {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    //if thread missing try to find on serv
    //if (target == null) return false;
    //final surl = target?.siteUrl.isNotEmpty ? target.siteUrl : _homeSite;

    final r = await homeSite?.refreshAuthorViaHTTP(shortLink);
    if (r == null) return r;
    _tbsCache.clear();
    return r;
  }

  Future<Uint8List> pullIPFScid(String cid, int xx) async {
    if (cid.isEmpty) {
      return Uint8List(0);
    }
    if (homeSite == null) return Uint8List(0);
    return homeSite!.xxC.pullIPFScid(cid, xx);
  }

  //unstuck ipfs queue
  int get ipfsQflushFailed => _sites[_homeSite]!.ipfsQflushFailed;

  ModeratorEntry? findModeratorEntryByShortLink(String shortLink) {
    for (var i in _sites.values) {
      final t = i.findModeratorEntryByShortLink(shortLink);
      if (t != null) return t;
    }
    return null;
  }

  Future<bool> get refreshAllViaHTTP async {
    homeSite?.refreshViaHTTP();

    return true;
  }

  List<int> get stats {
    final s = [
      0, //IOreceived
      0, //IOsent
      0,
      0,
      0, //memUsed
      0, //audience
      0, 0, 0, 0, 0, 0, 0
    ];
    //combined stats for each followed server
    allSites.forEach((element) async {
      var ii = 0;
      for (var i in element.stats) {
        s[ii] = s[ii] + i;
        ii++;
      }
    });
    return s;
  }

  bool online = true;
  void setOnline() {
    online = true;
    allSites.map((e) => e.setOnline);
  }

  void setOffline() {
    online = false;
    allSites.map((e) => e.setOffline);
  }

  /*Future<bool> switchHomeSite(String name, String uri) async {
    //if (siteKeys.contains(b)) {
    await homeSite?.logout;
    return switchSite(name, uri);
    //}
  }*/

  Uint8List get katamari {
    return homeSite != null ? homeSite!.katamari : Uint8List(0);
  }
}
