import 'dart:convert';
import 'dart:typed_data';

class SitesPreferencesEntry {
  String siteName;
  String siteUrl;
  String nick;
  String avatar;
  String ipfsGWup_a = '';
  String ipfsGWup_b = '';
  String ipfsGW = '';
  String dartSessionID = '';
  int lastAccess = 0;
  SitesPreferencesEntry(
      this.siteName, this.siteUrl, this.nick, this.avatar, this.lastAccess);
  Map toJson() => {
        'name': siteName,
        'url': siteUrl,
        'nick': nick,
        'avatar': avatar,
        'ipfsGWup_a': ipfsGWup_a,
        'ipfsGWup_b': ipfsGWup_b,
        'ipfsGW': ipfsGW,
        'dartSessionID': dartSessionID,
        'lastAccess': lastAccess
      };
  factory SitesPreferencesEntry.fromJson(dynamic json) {
    if (json['lastAccess'] == null) {
      return SitesPreferencesEntry(
          json['name'], json['url'], json['nick'], json['avatar'], 0);
    }
    return SitesPreferencesEntry(json['name'], json['url'], json['nick'],
        json['avatar'], json['lastAccess']);
  }
}

class SitesPreferences {
  List<dynamic> _sites = [];
  SitesPreferences(this._sites);

  List<dynamic> get sites => _sites;
  List<dynamic> get names => _sites.map((e) => e.siteName).toList();
  SitesPreferencesEntry? site(String siteName) {
    if (!names.contains(siteName)) return null;
    return _sites.where((e) => e.siteName == siteName).toList().first;
  }

  bool _dirty = false;
  bool get isDirty => _dirty;
  void unDirty() {
    _dirty = false;
  }

  SitesPreferencesEntry? get recentSite {
    if (names.isEmpty) return null;
    var ls = _sites.first;
    for (var i in _sites) {
      if (i.lastAccess > ls.lastAccess) {
        ls = i;
      }
    }
    return ls;
  }

  bool addSite(siteName, siteUrl, nick, avatar, lastAccess) {
    List<dynamic> _l =
        _sites.where((element) => element.siteName != siteName).toList();
    _l.add(SitesPreferencesEntry(siteName, siteUrl, nick, avatar, lastAccess));
    _sites = _l;
    _dirty = true;
    return true;
  }

  bool unfollowSite(siteName) {
    final n = _sites.map((e) => e.siteName);
    if (!n.contains(siteName)) {
      return false;
    }
    final rs = _sites.where((element) => element.siteName != siteName).toList();
    _sites = rs;
    _dirty = true;
    return true;
  }

  bool setNickAvatar(site, nick, avatar, lastAccess) {
    if (!names.contains(site)) {
      return false;
    }
    bool u = false;
    _sites.forEach((element) {
      if (element.siteName == site) {
        if (nick.isNotEmpty) element.nick = nick;
        if (avatar.isNotEmpty) element.avatar = avatar;
        element.lastAccess = lastAccess;
        _dirty = true;
      }
    });
    return _dirty;
  }

  String toJson() => jsonEncode(_sites);
  factory SitesPreferences.fromJson(data) {
    final ll = jsonDecode(data);

    final si = ll.map((e) => SitesPreferencesEntry.fromJson(e)).toList();
    return SitesPreferences(si);
  }
}
