import 'dart:async';
import 'dart:convert';
import 'dart:typed_data';

import 'package:http/http.dart' as http;
import 'package:interzone/interop/moderatorCapsuleEntry.dart';

class httpInterzone {
  String uri = "";
  String cookie = '';
  Uint8List _katamari = Uint8List(0);
  String dartSessionId = '';
  bool _busy = false;
  bool _handshaked = false;

  bool get isBusy => _busy;
  bool online = true;

  final _timoutIn = Duration(milliseconds: 2912);
  void setOnline() {
    online = true;
  }

  void setOffline() {
    online = false;
  }

  httpInterzone();
  Uint8List xxWantsList = Uint8List(0);

  Future<Uint8List> handshake(String server, Uint8List katamari) async {
    if (_handshaked && dartSessionId.isNotEmpty && katamari.isNotEmpty) {
      return refresh(server, katamari);
    }
    uri = server;
    var url = Uri.parse(server + '?bin=true');
    //var response =
    //    await http.post(url, body: {'name': 'doodle', 'color': 'blue'});
    var response;
    try {
      if (dartSessionId.isNotEmpty) {
        response = await http.get(url, headers: {
          'Cookie': 'DARTSESSID=' + dartSessionId
        }).timeout(_timoutIn);
      } else {
        response = await http.get(url);
      }
    } catch (e) {
      if (!(e is TimeoutException)) setOffline();
      return Uint8List(0);
    }
    if (response == null) return Uint8List(0);
    //data is flowing, we are good
    setOnline();

    if (response.statusCode != 200) {
      return Uint8List(0);
    }

    print('Response status: ${response.statusCode}');
    String? h = response.headers["set-cookie"];
    if (h != null) {
      cookie = h.split(' ').first;
      String dSid = cookie.split('=').last;
      dartSessionId = dSid.substring(0, dSid.length - 1);
      _handshaked = true;
    } else {
      print('cookie header missing!!');
    }
    String? hw = response.headers["xxwants"];
    if (hw != null) {
      xxWantsList = base64Decode(hw);
    } else {
      //empty list
      xxWantsList = Uint8List(0);
    }
    print('dartSessionID' + dartSessionId);
    print(response.contentLength);

    return response.bodyBytes;
  }

  Future<Uint8List> newReply(
      String server,
      String text,
      String rootShortlink,
      String ppShortlink,
      Uint8List katamari,
      String nick,
      String avatar) async {
    var url = Uri.parse(server + '/reply/' + rootShortlink + '?bin=true');
    print('replyThreadThread:' + text);
    try {
      var response = await http.post(
        url,
        body: {
          'reply': text,
          'katamari': base64Encode(katamari),
          'nick': nick,
          'avatar': avatar,
          'pp': ppShortlink
        },
        headers: {'Cookie': 'DARTSESSID=' + dartSessionId},
      ).timeout(_timoutIn);
      if (response == null) return Uint8List(0);
      setOnline();
      if (response.statusCode == 409) {
        //dup
        return Uint8List(1);
      }
      if (response.statusCode == 502) {
        //dup
        return Uint8List(2);
      }
      if (response.statusCode != 200) return Uint8List(0);

      print('Response status: ${response.statusCode}');
      String? h = response.headers["set-cookie"];

      if (h != null) {
        cookie = h.split(' ').first;
        String dSid = cookie.split('=').last;
        dartSessionId = dSid.substring(0, dSid.length - 1);
      }

      print(cookie);
      print(response.contentLength);
      return response.bodyBytes;
    } catch (e) {
      print('Future<Uint8List> newReply httpconnector ERROR');
      if (!(e is TimeoutException)) setOffline();
      print(e);
      return Uint8List(0);
    }
  }

  Future<Uint8List> newThread(String _baseUri, String text, String preEntry,
      Uint8List katamari, String nick, String avatar) async {
    var url = Uri.parse(_baseUri + '/new/' + preEntry + '?bin=true');
    print('newThread:' + text);
    var response;
    try {
      response = await http.post(url, body: {
        'post': text,
        'katamari': base64Encode(katamari),
        'nick': nick,
        'avatar': avatar
      }, headers: {
        'Cookie': 'DARTSESSID=' + dartSessionId
      }).timeout(_timoutIn);
    } catch (e) {
      if (!(e is TimeoutException)) setOffline();
      return Uint8List(0);
    }
    if (response == null) return Uint8List(0);
    setOnline();
    if (response.statusCode == 409) {
      //dup
      return Uint8List(1);
    }
    if (response.statusCode == 502) {
      //dup
      return Uint8List(2);
    }
    if (response.statusCode != 200) return Uint8List(0);

    print('Response status: ${response.statusCode}');
    String? h = response.headers["set-cookie"];

    if (h != null) {
      cookie = h.split(' ').first;
      String dSid = cookie.split('=').last;
      dartSessionId = dSid.substring(0, dSid.length - 1);
    }

    print(cookie);
    print(response.contentLength);
    return response.bodyBytes;
  }

  Future<Uint8List> newThreadWithIPFSAttachement(
      String _baseUri,
      String text,
      String preEntry,
      Uint8List katamari,
      String nick,
      int atl,
      int xxAt,
      String cid,
      String avatar,
      int atType) async {
    var url = Uri.parse(_baseUri + '/new/' + preEntry + '?bin=true');

    final ModeratorCapsuleItemAttachment att =
        ModeratorCapsuleItemAttachment(atl, xxAt, cid);
    var spicedCid = ' !&' + cid;
    //0 image,1 video,2 audio
    if (atType == 1) spicedCid = ' !∞' + cid;
    if (atType == 2) spicedCid = ' !£' + cid;

    var request = http.MultipartRequest("POST", url)
      ..fields['katamari'] = base64Encode(katamari)
      ..fields['post'] = text + spicedCid
      ..fields['nick'] = nick
      ..fields['avatar'] = avatar
      ..fields['attachment'] = base64Encode(att.binary);

    request.headers.addAll({'Cookie': 'DARTSESSID=' + dartSessionId});
    var response;
    try {
      response = await request.send().timeout(_timoutIn);
    } catch (e) {
      if (!(e is TimeoutException)) setOffline();
      return Uint8List(0);
    }
    setOnline();
    if (response.statusCode == 409) {
      //dup
      return Uint8List(1);
    }
    if (response.statusCode == 502) {
      //dup
      return Uint8List(2);
    }

    if (response.statusCode == 200) {
      print("Image Uploaded");
    } else {
      print("Upload Failed");
      return Uint8List(0);
    }
    String? h = response.headers["set-cookie"];
    if (h != null) {
      cookie = h.split(' ').first;
      String dSid = cookie.split('=').last;
      dartSessionId = dSid.substring(0, dSid.length - 1);
    }
    print(cookie);
    print(response.contentLength);
    return response.stream.toBytes();
  }

  Future<Uint8List> newReplyWithIPFSAttachement(
      String _baseUri,
      String rootShortLink,
      String text,
      String preEntry,
      Uint8List katamari,
      String nick,
      int atl,
      int xxAt,
      String cid,
      String avatar,
      int atType) async {
    var url = Uri.parse(_baseUri + '/reply/' + rootShortLink + '?bin=true');

    final ModeratorCapsuleItemAttachment att =
        ModeratorCapsuleItemAttachment(atl, xxAt, cid);

    var spicedCid = ' !&' + cid;
    //0 image,1 video,2 audio
    if (atType == 1) spicedCid = ' !∞' + cid;
    if (atType == 2) spicedCid = ' !£' + cid;

    var request = http.MultipartRequest("POST", url)
      ..fields['katamari'] = base64Encode(katamari)
      ..fields['reply'] = text + spicedCid
      ..fields['nick'] = nick
      ..fields['avatar'] = avatar
      ..fields['attachment'] = base64Encode(att.binary);

    request.headers.addAll({'Cookie': 'DARTSESSID=' + dartSessionId});
    var response;
    try {
      response = await request.send().timeout(_timoutIn);
    } catch (e) {
      if (!(e is TimeoutException)) setOffline();
      return Uint8List(0);
    }
    setOnline();
    if (response.statusCode == 409) {
      //dup
      return Uint8List(1);
    }
    if (response.statusCode == 502) {
      //dup
      return Uint8List(2);
    }
    if (response.statusCode == 200) {
      print("Image Uploaded");
    } else {
      print("Upload Failed");
      return Uint8List(0);
    }
    String? h = response.headers["set-cookie"];
    if (h != null) {
      cookie = h.split(' ').first;
      String dSid = cookie.split('=').last;
      dartSessionId = dSid.substring(0, dSid.length - 1);
    }
    print(cookie);
    print(response.contentLength);
    return response.stream.toBytes();
  }

  Future<Uint8List> newThreadWithAttachement(
      String siteUrl,
      String text,
      Uint8List attachement,
      String preEntry,
      Uint8List katamari,
      String nick,
      int atl,
      int xxAt,
      String cid,
      String avatar) async {
    var url = Uri.parse(siteUrl + '/new/' + preEntry + '?bin=true');
    final ModeratorCapsuleItemAttachment att =
        ModeratorCapsuleItemAttachment(atl, xxAt, cid);

    var request = http.MultipartRequest("POST", url)
      ..fields['katamari'] = base64Encode(katamari)
      ..fields['post'] = text
      ..fields['nick'] = nick
      ..fields['avatar'] = avatar
      ..fields['attachment'] = base64Encode(att.binary);

    /*var multipartFile = http.MultipartFile.fromBytes(
        'filename', attachement); //("image", attachement, length,
    request.files.add(multipartFile);*/

    request.headers.addAll({'Cookie': 'DARTSESSID=' + dartSessionId});
    var response;
    try {
      response = await request.send().timeout(_timoutIn);
    } catch (e) {
      if (!(e is TimeoutException)) setOffline();
      return Uint8List(0);
    }
    setOnline();
    if (response.statusCode == 409) {
      //dup
      return Uint8List(1);
    }

    if (response.statusCode == 200) {
      print("Image Uploaded");
    } else {
      print("Upload Failed");
    }
    String? h = response.headers["set-cookie"];
    if (h != null) {
      cookie = h.split(' ').first;
      String dSid = cookie.split('=').last;
      dartSessionId = dSid.substring(0, dSid.length - 1);
    }
    print(cookie);
    print(response.contentLength);
    return response.stream.toBytes();
  }

  Future<bool> getThreads() async {
    return false;
  }

  Future<bool> getThread() async {
    return false;
  }

  Future<Uint8List> refresh(String server, Uint8List katamari) async {
    if (!_handshaked && katamari.isEmpty)
      return handshake(server, Uint8List(0));
    uri = server;

    var url = Uri.parse(uri + '/?bin=true');
    //var response =
    //    await http.post(url, body: {'name': 'doodle', 'color': 'blue'});
    var response;
    try {
      response = await http.post(url, body: {
        'katamari': base64Encode(katamari)
      }, headers: {
        'Cookie': 'DARTSESSID=' + dartSessionId
      }).timeout(_timoutIn);
    } catch (e) {
      if (!(e is TimeoutException)) setOffline();
      return Uint8List(0);
    }
    if (response == null) return Uint8List(0);
    setOnline();
    if (response.statusCode != 200) return Uint8List(0);
    print('Response status: ${response.statusCode}');
    String? h = response.headers["set-cookie"];
    if (h != null) {
      cookie = h.split(' ').first;
      String dSid = cookie.split('=').last;
      dartSessionId = dSid.substring(0, dSid.length - 1);
    }
    print(cookie);
    print(response.contentLength);
    String? hw = response.headers["xxwants"];
    if (hw != null) {
      xxWantsList = base64Decode(hw);
    } else {
      //empty list
      xxWantsList = Uint8List(0);
    }
    return response.bodyBytes;
  }

  Future<Uint8List> refreshTagged(String server, String shortlink) async {
    if (!_handshaked && _katamari.isEmpty) return handshake(server, _katamari);
    uri = server;

    var url = Uri.parse(uri + '/tag/' + shortlink + '?bin=true');
    //var response =
    //    await http.post(url, body: {'name': 'doodle', 'color': 'blue'});
    var response;
    try {
      response = await http.post(url, body: {
        'katamari': base64Encode(_katamari)
      }, headers: {
        'Cookie': 'DARTSESSID=' + dartSessionId
      }).timeout(_timoutIn);
    } catch (e) {
      if (!(e is TimeoutException)) setOffline();
      return Uint8List(0);
    }
    if (response == null) return Uint8List(0);
    setOnline();
    if (response.statusCode != 200) return Uint8List(0);
    print('Response status: ${response.statusCode}');
    String? h = response.headers["set-cookie"];
    if (h != null) {
      cookie = h.split(' ').first;
      String dSid = cookie.split('=').last;
      dartSessionId = dSid.substring(0, dSid.length - 1);
    }
    print(cookie);
    print(response.contentLength);
    String? hw = response.headers["xxwants"];
    if (hw != null) {
      xxWantsList = base64Decode(hw);
    } else {
      //empty list
      xxWantsList = Uint8List(0);
    }
    return response.bodyBytes;
  }

  Future<Uint8List> refreshAuthor(String server, String shortlink) async {
    if (!_handshaked && _katamari.isEmpty) return handshake(server, _katamari);

    uri = server;

    var url = Uri.parse(uri + '/a/' + shortlink + '?bin=true');
    //var response =
    //    await http.post(url, body: {'name': 'doodle', 'color': 'blue'});

    var response;
    try {
      response = await http.post(url, body: {
        'katamari': base64Encode(_katamari)
      }, headers: {
        'Cookie': 'DARTSESSID=' + dartSessionId
      }).timeout(_timoutIn);
    } catch (e) {
      if (!(e is TimeoutException)) setOffline();
      return Uint8List(0);
    }

    if (response == null) return Uint8List(0);
    setOnline();
    if (response.statusCode != 200) return Uint8List(0);
    print('Response status: ${response.statusCode}');
    String? h = response.headers["set-cookie"];
    if (h != null) {
      cookie = h.split(' ').first;
      String dSid = cookie.split('=').last;
      dartSessionId = dSid.substring(0, dSid.length - 1);
    }
    print(cookie);
    print(response.contentLength);
    String? hw = response.headers["xxwants"];
    if (hw != null) {
      xxWantsList = base64Decode(hw);
    } else {
      //empty list
      xxWantsList = Uint8List(0);
    }
    return response.bodyBytes;
  }

  Future<Uint8List> refreshThread(String server, String shortlink) async {
    if (!_handshaked && _katamari.isEmpty) return handshake(server, _katamari);

    uri = server;

    var url = Uri.parse(uri + '/t/' + shortlink + '?bin=true');
    //var response =
    //    await http.post(url, body: {'name': 'doodle', 'color': 'blue'});

    var response;
    try {
      response = await http.post(url, body: {
        'katamari': base64Encode(_katamari)
      }, headers: {
        'Cookie': 'DARTSESSID=' + dartSessionId
      }).timeout(_timoutIn);
    } catch (e) {
      if (!(e is TimeoutException)) setOffline();
      return Uint8List(0);
    }

    //var response =
    //    await http.get(url, headers: {'Cookie': 'DARTSESSID=' + dartSessionId});
    if (response == null) return Uint8List(0);
    setOnline();
    if (response.statusCode != 200) return Uint8List(0);
    print('Response status: ${response.statusCode}');
    String? h = response.headers["set-cookie"];
    if (h != null) {
      cookie = h.split(' ').first;
      String dSid = cookie.split('=').last;
      dartSessionId = dSid.substring(0, dSid.length - 1);
    }
    print(cookie);
    print(response.contentLength);
    String? hw = response.headers["xxwants"];
    if (hw != null) {
      xxWantsList = base64Decode(hw);
    } else {
      //empty list
      xxWantsList = Uint8List(0);
    }
    return response.bodyBytes;
  }

  bool _satisfyingXX = false;

  Future<bool> satisfyXXwant(Uint8List attachement) async {
    if (_satisfyingXX) return false;
    _satisfyingXX = true;
    var url = Uri.parse(uri + '/xx');
    var request = http.MultipartRequest("POST", url);
    var multipartFile = http.MultipartFile.fromBytes(
        'xxsatisfy', attachement); //("image", attachement, length,
    request.files.add(multipartFile);
    request.headers.addAll({'Cookie': 'DARTSESSID=' + dartSessionId});
    var response = await request.send().timeout(_timoutIn);
    _satisfyingXX = false;
    if (response.statusCode == 200) {
      print("xxSatisfied");
      return true;
    } else {
      print("Upload Failed");
    }

    print(cookie);
    print(response.contentLength);
    return false;
  }

  Future<Uint8List> newDrop(String rootShortlink, Uint8List katamari) async {
    var url = Uri.parse(uri + '/drop/' + rootShortlink + '?bin=true');
    try {
      var response = await http.post(url, body: {
        'katamari': base64Encode(katamari),
      }, headers: {
        'Cookie': 'DARTSESSID=' + dartSessionId
      }).timeout(_timoutIn);
      if (response == null) return Uint8List(0);
      if (response.statusCode != 200) return Uint8List(0);

      print('Response status: ${response.statusCode}');
      String? h = response.headers["set-cookie"];

      if (h != null) {
        cookie = h.split(' ').first;
        String dSid = cookie.split('=').last;
        dartSessionId = dSid.substring(0, dSid.length - 1);
      }

      return response.bodyBytes;
    } catch (e) {
      print('Future<Uint8List> newDrop httpconnector ERROR');
      print(e);
      if (!(e is TimeoutException)) setOffline();
      return Uint8List(0);
    }
  }

  Future<Uint8List> newMute(String rootShortlink, Uint8List katamari) async {
    var url = Uri.parse(uri + '/mute/' + rootShortlink + '?bin=true');
    try {
      var response = await http.post(url, body: {
        'katamari': base64Encode(katamari),
      }, headers: {
        'Cookie': 'DARTSESSID=' + dartSessionId
      }).timeout(_timoutIn);
      if (response == null) return Uint8List(0);
      if (response.statusCode != 200) return Uint8List(0);

      print('Response status: ${response.statusCode}');
      String? h = response.headers["set-cookie"];

      if (h != null) {
        cookie = h.split(' ').first;
        String dSid = cookie.split('=').last;
        dartSessionId = dSid.substring(0, dSid.length - 1);
      }

      return response.bodyBytes;
    } catch (e) {
      print('Future<Uint8List> newDrop httpconnector ERROR');
      print(e);
      if (!(e is TimeoutException)) setOffline();
      return Uint8List(0);
    }
  }

  Future<Uint8List> newLike(String rootShortlink, Uint8List katamari) async {
    var url = Uri.parse(uri + '/like/' + rootShortlink + '?bin=true');
    try {
      var response = await http.post(url, body: {
        'katamari': base64Encode(katamari),
      }, headers: {
        'Cookie': 'DARTSESSID=' + dartSessionId
      }).timeout(_timoutIn);
      if (response == null) return Uint8List(0);
      if (response.statusCode != 200) return Uint8List(0);

      print('Response status: ${response.statusCode}');
      String? h = response.headers["set-cookie"];

      if (h != null) {
        cookie = h.split(' ').first;
        String dSid = cookie.split('=').last;
        dartSessionId = dSid.substring(0, dSid.length - 1);
      }

      return response.bodyBytes;
    } catch (e) {
      print('Future<Uint8List> newLike httpconnector ERROR');
      print(e);
      if (!(e is TimeoutException)) setOffline();
      return Uint8List(0);
    }
  }

  Future<Uint8List> newShare(String rootShortlink, Uint8List katamari) async {
    var url = Uri.parse(uri + '/share/' + rootShortlink + '?bin=true');
    try {
      var response = await http.post(url, body: {
        'katamari': base64Encode(katamari),
      }, headers: {
        'Cookie': 'DARTSESSID=' + dartSessionId
      }).timeout(_timoutIn);
      if (response == null) return Uint8List(0);
      //504 gateway tim out!
      if (response.statusCode != 200) return Uint8List(0);

      print('Response status: ${response.statusCode}');
      String? h = response.headers["set-cookie"];

      if (h != null) {
        cookie = h.split(' ').first;
        String dSid = cookie.split('=').last;
        dartSessionId = dSid.substring(0, dSid.length - 1);
      }

      return response.bodyBytes;
    } catch (e) {
      print('Future<Uint8List> newShare httpconnector ERROR');
      print(e);
      if (!(e is TimeoutException)) setOffline();
      return Uint8List(0);
    }
  }
}
