import 'dart:async';
import 'dart:typed_data';
import 'interop/xxhash/lib/fast_hash.dart';
import 'package:http/http.dart' as http;

class IpfsQueue {
  bool _ifpsDead = false;
  Map<int, String> _inserted = {};
  Map<int, Uint8List> _queued = {};
  bool online = true;
  void setOnline() {
    online = true;
    //reset ipfs state in case it was bad connection
    _ifpsDead = false;
    _failed.clear();
    //keep sending stuff
    push();
  }

  void setOffline() {
    online = false;
  }

  Set<int> _failed = {};
  int get flushFailed {
    final f = _failed.length;
    _failed.clear();
    return f;
  }

  String insertToQueue(int xxHash, Uint8List attachement) {
    if (_inserted[xxHash] != null) return _inserted[xxHash]!;
    if (_queued[xxHash] != null) {
      return '';
    }
    _queued[xxHash] = attachement;
    return '';
  }

  String _endpointUrl = '';
  String _pinata_api_key = '';
  String _pinata_secret_api_key = '';

  setIpfsCredentials(
      String pinata_api_key, String pinata_secret_api_key, String endpointUrl) {
    _endpointUrl = endpointUrl;
    _pinata_api_key = pinata_api_key;
    _pinata_secret_api_key = pinata_secret_api_key;
    return true;
  }

  bool _pushingToIpfs = false;
  String findCid(int xx) {
    if (_inserted[xx] == null) return '';
    return _inserted[xx]!;
  }

  Future<bool> push() async {
    if (_queued.isEmpty) return false;
    if (_pushingToIpfs) return false;
    if (!online) return false;
    _pushingToIpfs = true;

    final goe = _queued.keys
        .where((element) =>
            !!_inserted.containsKey(element) && !!_failed.contains(element))
        .toList();
    if (goe.isEmpty) return false;

    final xdat = _queued[goe.first];
    final xx = goe.first;

    return await pushToPinataIPFS(
            _pinata_api_key, _pinata_secret_api_key, xdat!)
        .then((value) {
      if (value.isNotEmpty) {
        _queued[xx] != null;
        _pushingToIpfs = false;
        return true;
      }
      _failed.add(xx);
      _pushingToIpfs = false;
      return false;
    });
  }

  Future<String> pushToPinataIPFS(String pinata_api_key,
      String pinata_secret_api_key, Uint8List data) async {
    if (pinata_api_key.isEmpty) {
      pinata_api_key = _pinata_api_key;
      pinata_secret_api_key = _pinata_secret_api_key;
    }
    final String ur = 'https://api.pinata.cloud/pinning/pinFileToIPFS';
    var url = Uri.parse(ur);
    var request = http.MultipartRequest("POST", url);
    var multipartFile = http.MultipartFile.fromBytes('file', data,
        filename: 'ada'); //("image", attachement, length,
    request.files.add(multipartFile);
    request.headers.addAll({
      'pinata_api_key': pinata_api_key,
      'pinata_secret_api_key': pinata_secret_api_key
    });
    var r = '';
    try {
      r = await request.send().then((r) async {
        if (r == null) return '';
        setOnline();
        if (r.statusCode == 403) {
          print('FORBIDDEN: ipfs keys borked. panic.');
          _ifpsDead = true;
          return '';
        }
        if (r.statusCode != 200) return '';
        final responseString = await r.stream.bytesToString();

        if (responseString.startsWith('{"IpfsHash":')) {
          final f = responseString.split('"');
          return (f[3]);
        }
        return '';
      });
    } catch (e) {
      if (!(e is TimeoutException)) setOffline();
      return '';
    }
    return r;
  }

  int xxHash(Uint8List b) {
    if (b.isEmpty) return -1;
    //return b.hashCode;
    final digest = xxHash64.convert(b);
    return digest.x.toInt();
  }
}
