import 'dart:async';
import 'dart:io';
import 'dart:isolate';
import 'dart:typed_data';
import 'package:crypto/crypto.dart';
import 'package:image_compression/image_compression.dart';

import 'http/http_connector.dart';

import 'moderator/OutgoingMessageCache.dart';
import 'moderator/moderatorCore.dart';
import 'moderator/OmniRemembers.dart';
import 'moderator/ogParse.dart';

import 'disk.dart';
export 'moderator/moderatorCore.dart';
export 'disk.dart';
import 'package:hex/hex.dart';
import 'interop/xxhash/lib/fast_hash.dart';
import 'package:blurhash_dart/blurhash_dart.dart';
import 'package:image/image.dart' as img;
import 'ipfsQueue.dart';
export 'disk.dart';
import 'package:mime/mime.dart';

class world {
  httpInterzone? _io;
  interzoneDisk? _disk;
  OmniRemembers _omni = OmniRemembers();
  ModeratorEntrySet data = ModeratorEntrySet();
  ModeratorCapsule? _latestCapsule;
  Uint8List _katamari = Uint8List(0);
  Uint8List get katamari => _katamari;

  xxCache _xxCache = xxCache();
  Future<bool> satisfyXX(int xx, Uint8List data) =>
      _xxCache.satisfy(xx, data, false);
  OutgoingMessageCache _ogMessages = OutgoingMessageCache();
  String _server = '';
  String _siteName = '';
  String get siteName => _siteName;
  String get baseUrl => _server;
  String _nicke = '';
  String _avatare = '🐷';
  String get currentNick => _nicke;
  String get currentAvatar => _avatare;
  String appSplashImage = '';
  bool _dirty = false;
  IpfsQueue _ipfsQ = IpfsQueue();
  bool _terminated = false;
  int dataIn = 0;
  int dataOut = 0;
  bool get _online => _io != null ? _io!.online : false;
  void setOnline() {
    _io?.setOnline;
    _ipfsQ.setOnline;
    _xxCache.setOnline();
  }

  void setOffline() {
    _io?.setOffline;
    _ipfsQ.setOffline;
    _xxCache.setOffline();
  }

  void unDirty() {
    //call this from outside only
    _dirty = false;
  }

  bool get isDirty => _dirty;
  world();

  xxCache get xxC => _xxCache;
  Uint8List get xxWants => _xxCache.wantsBinary;
  //action handlers
  dynamic messageActionSignalHandler;
  Set<String> _mutedAuthors = {};
  Set<int> _deletedEntries = {};
  void terminate() {
    _terminated = true;
  }

  int get now => DateTime.now().millisecondsSinceEpoch;
  void ping() {
    refreshViaHTTP();
  }

  void defocus() {
    //TODO
  }

  int get dataPulledDisk => _disk!.dataPulledDisk;
  int get dataPushedDisk => _disk!.dataPushedDisk;

  ModeratorEntrySet get messages {
    if (_mutedAuthors.isEmpty && _deletedEntries.isEmpty) return data;
    ModeratorEntrySet _l = ModeratorEntrySet();
    data.all.forEach((element) {
      if (_mutedAuthors.contains(element.author)) return;
      if (_deletedEntries.contains(element.keyHint)) return;
      element.flags.isSeen = _omni.isSeen(element.hint);
      element.flags.isLiked = _omni.isLiked(element.hint);
      _l.add(element);
    });
    return _l;
  }

  ModeratorEntrySet get LatestThreads {
    if (_mutedAuthors.isEmpty && _deletedEntries.isEmpty) return data;
    ModeratorEntrySet _l = ModeratorEntrySet();
    data.latestThreads.forEach((element) {
      if (!element.flags.isPost) return;
      if (_mutedAuthors.contains(element.author)) return;
      if (_deletedEntries.contains(element.keyHint)) return;
      element.flags.isSeen = _omni.isSeen(element.hint);
      element.flags.isLiked = _omni.isLiked(element.hint);

      _l.add(element);
    });
    return _l;
  }

  List<ModeratorEntry> threadByShortLink(String shortLink) {
    List<ModeratorEntry> _ll = [];
    final ll2 = data.threadByShortLink(shortLink);
    if (ll2.isEmpty) return _ll;
    bool pkc = false;
    for (var element in ll2) {
      if (_deletedEntries.contains(element.keyHint)) continue;
      if (element.flags.isPost && !pkc) {
        pkc = true;
        bool newa = false;
        List<ModeratorEntry> _lll = [];
        ll2.forEach((ce) {
          if (_deletedEntries.contains(ce.keyHint)) return;
          if (ce.flags.isReply) {
            ce.flags.isSeen = _omni.isSeen(ce.hint);
            ce.flags.isLiked = _omni.isLiked(ce.hint);
            //any single item unseen makes parent unseen
            if (!ce.flags.isSeen) newa = true;
            _lll.add(ce);
          }
        });
        if (newa) {
          element.flags.isSeen = false;
        } else {
          element.flags.isSeen = _omni.isSeen(element.hint);
        }
        _ll.add(element);
        if (_lll.isNotEmpty) _ll.addAll(_lll);
        break;
      }
    }

    return _ll;
  }

  List<ModeratorEntry> unreadThreadByShortLink(String shortLink) {
    //OMNI here to track seen unseen / seen
    List<ModeratorEntry> _ll = [];
    for (var element in data.threadByShortLink(shortLink)) {
      if (element.shortLink == shortLink) {
        bool newa = false;
        List<ModeratorEntry> _lll = [];
        data.childrenOf(element.hint).forEach((ce) {
          if (ce.flags.isReply) {
            ce.flags.isSeen = _omni.isSeen(ce.hint);
            //any single item unseen makes parent unseen
            if (!ce.flags.isSeen) newa = true;
            _lll.add(ce);
          }
        });
        if (newa) {
          element.flags.isSeen = false;
        } else {
          element.flags.isSeen = _omni.isSeen(element.hint);
        }
        _ll.add(element);
        if (_lll.isNotEmpty) _ll.addAll(_lll);
      }
    }
    return _ll;
  }

  bool isSeenOmni(String shortLink) {
    ModeratorEntry? m = data.read(shortLink);
    if (m == null) return false;
    return _omni.isSeen(m.hint);
  }

  bool markSeenOmni(String shortLink) {
    //set this entry as seen
    ModeratorEntry? m = data.read(shortLink);
    if (m == null) return false;
    if (_omni.isSeen(m.hint)) return true;
    //refresh for cache
    _dirty = true;
    return _omni.markSeen(m.hint);
  }

  bool unPeekEntryOmni(String shortLink) {
    //set this entry as seen
    ModeratorEntry? m = data.read(shortLink);
    if (m == null) return false;
    return _omni.unPeekEntry(m.hint);
  }

  bool peekThreadOmni(String shortLink) {
    //mark whole thread as seen
    return true;
  }

  bool dropEntryOmni(String shortLink) {
    //set this entry as seen
    return true;
  }

  bool dropThreadOmni(String shortLink) {
    //mark whole thread dropped
    return true;
  }

  Future<String> xxHashHex(Uint8List data) async {
    final attXX = data.hashCode;
    //final t = await xxHash(data).bake();
    final bdata = Uint8List(8)
      ..buffer.asByteData().setInt64(0, attXX, Endian.big);
    return HEX.encode(bdata);
  }

  List<ModeratorEntry>? updateLatestCapsule(Uint8List bin) {
    //throw latest shit in the entrySet
    //TODO check for exact previous
    //error codes return less data
    //409,502
    if (bin.lengthInBytes < 28) return [];
    print('updateLatestCapsule len');
    print(bin.lengthInBytes);
    dataIn += bin.lengthInBytes;
    var mc = ModeratorCapsule.fromBinary(bin);
    //if (mc.isDEADBEEF) return null;
    if (mc == null) return null;
    _latestCapsule = mc;
    _katamari = mc.katamari;
    //writes new shit to disk too
    final t = insertToLocalEntrySet(mc);
    if (t.isNotEmpty) {
      _dirty = true;
      if (mc.katamari.isNotEmpty) _disk?.commitKatamari(mc.katamari);
    }
    return t;
  }

  Future<List<ModeratorEntry>?> init(String _siteNamen, String server,
      String author, String avatar, interzoneDisk _disco) async {
    _server = server;
    _siteName = _siteNamen;
    _io = httpInterzone();

    _disk = _disco;
    _ogMessages.init(_disco, _siteName);
    _omni.init(_disco, _siteName);

    //await _disk.init('interzoneData', false, _disk);
    String previousDartSeshId = '';
    // await _disk.readPreferences;

    /*List<String>? _prifs = await _disk?.readPreferences.then((_prifs) {
      if (_prifs != null) {
        previousDartSeshId = _prifs[0];
        if (_prifs.length > 1) {
          _nicke = _prifs[1];
          if (_prifs.length > 2) {
            _avatare = _prifs[2];
          }
        }
      } else {
        _nicke = author;
        _avatare = avatar;
      }
    }); */

    try {
      await _disk?.readIpfsPreferences.then((value) {
        if (value != null) {
          if (value.isNotEmpty) {
            final cd = value.split(':');
            _ipfsQ.setIpfsCredentials(cd[0], cd[1], '');
          }
        }
      });
    } catch (e) {
      print('world::readDeletesPreferences exploded');
      print(e);
    }

    try {
      _disk?.readDeletesPreferences.then((value) => {
            if (value != null)
              {
                if (value.isNotEmpty) {_deletedEntries = value.toSet()}
              }
          });
    } catch (e) {
      print('world::readDeletesPreferences exploded');
      print(e);
    }
    try {
      _disk?.readMutesPreferences.then((value) => {
            if (value != null)
              {
                if (value.isNotEmpty)
                  {
                    _mutedAuthors =
                        value.where((element) => element.isNotEmpty).toSet()
                  }
              }
          });
    } catch (e) {
      print('world::readMutesPreferences exploded');
      print(e);
    }

    _omni.init(_disk!, _siteName);

    await _xxCache.init(_disk!);
    _disk!.readCapsule.then((cps) {
      if (cps != null) {
        LocalEntrySetPrime(cps);
      }
    });

    Uint8List katt = await _disk!.readKatamari;
    if (katt.isNotEmpty) {
      _katamari = katt;
    }
    //async hundkarusellen!
    if (previousDartSeshId.isNotEmpty) {
      print('injected previous dart sesh id to help the police');
      _io!.dartSessionId = previousDartSeshId;
    }

    return login;
  }

  Future<List<ModeratorEntry>?> refreshViaHTTP() async {
    if (_io == null) return null;
    return await _io!.refresh(_server, _katamari).then((hs) async {
      if (hs.isNotEmpty) {
        dataIn += hs.lengthInBytes;
        //await _disk.commitPreferences(_io!.dartSessionId, nick);
        satisfyXXwant(_io!.xxWantsList);
        outgoingMessageCacheSend(0);
        return updateLatestCapsule(hs);
      }
      return null;
    });
  }

  Future<List<ModeratorEntry>?> refreshTaggedViaHTTP(String shortLink) async {
    if (_io == null) return null;
    return await _io!.refreshTagged(_server, shortLink).then((hs) async {
      if (hs.isNotEmpty) {
        dataIn += hs.lengthInBytes;
        return updateLatestCapsule(hs);
      }
    });
  }

  Future<List<ModeratorEntry>?> refreshAuthorViaHTTP(String shortLink) async {
    if (_io == null) return null;
    return await _io!.refreshAuthor(_server, shortLink).then((hs) async {
      if (hs.isNotEmpty) {
        dataIn += hs.lengthInBytes;
        return updateLatestCapsule(hs);
      }
    });
  }

  Future<List<ModeratorEntry>?> refreshThreadViaHTTP(String shortlink) async {
    if (_io == null) return null;
    return await _io!.refreshThread(_server, shortlink).then((hs) async {
      if (hs.isNotEmpty) {
        dataIn += hs.lengthInBytes;
        //await _disk.commitPreferences(_io!.dartSessionId, nick);
        satisfyXXwant(_io!.xxWantsList);
        return updateLatestCapsule(hs);
      }
      return null;
    });
  }

  Future<List<ModeratorEntry>?> refreshGroupThreadViaHTTP(
      String shortlink) async {
    if (_io == null) return null;
    return await _io!.refreshThread(_server, shortlink).then((hs) async {
      if (hs.isNotEmpty) {
        dataIn += hs.lengthInBytes;
        //await _disk.commitPreferences(_io!.dartSessionId, nick);
        satisfyXXwant(_io!.xxWantsList);
        return updateLatestCapsule(hs);
      }
      return null;
    });
  }

  Future<bool> setUserNickAndImage(
      String _nick, String _avatar, Uint8List attachement) async {
    //await _disk?.commitPreferences(_io!.dartSessionId, _nick, _avatar);
    return true;
  }

  Future<bool> setIpfsCredentials(String k1, String k2, String url) async {
    //await _disk?.commitIpfsPreferences(k1 + ':' + k2);
    return _ipfsQ.setIpfsCredentials(k1, k2, url);
  }

  int get ipfsQflushFailed => _ipfsQ.flushFailed;
  Future<int> newThreadToQueue(
      String _baseUrl,
      String text,
      Uint8List attachement,
      String _nick,
      String _avatar,
      String shortLinkOfPrevious,
      String cid,
      String attachmentType) async {
    if (shortLinkOfPrevious.isEmpty) {
      //patch in latest root
      final own = LatestThreads.latestMineShortLink;
      if (own.isEmpty) {
        shortLinkOfPrevious = LatestThreads.posts.last.shortLink;
      } else {
        shortLinkOfPrevious = own;
      }
    }

    final ogm = _ogMessages.addPost(_nick, _avatar, text, shortLinkOfPrevious,
        '', now, attachement, cid, attachmentType);

    if (ogm == null) return -1;
    if (!_online) {
      //just push to disk when no net availablus
      _ogMessages.pushToDiskQueue(ogm.timestamp);
    }
    return ogm.timestamp;
  }

  Map<int, int> _xxhc = {};
  Future<List<ModeratorEntry>?> newThread(
      String _baseUrl,
      String text,
      Uint8List attachement,
      String _nick,
      String _avatar,
      String shortLinkOfPrevious,
      String DEPcid,
      String attachmentType) async {
    //if (_latestCapsule == null) return null;
    if (_nick.isEmpty) {
      _nick = currentNick;
    }
    if (shortLinkOfPrevious.isEmpty) {
      shortLinkOfPrevious =
          messages.all.where((element) => element.flags.isPost).first.shortLink;
    }
    Uint8List nt;
    //assume net is live here
    if (text.contains('https://')) {
      bool bcHash = text.contains('#bandcamp');
      bool tagBc = false;
      if (!text.startsWith('[')) {
        final ali = text.split('https://');
        final alib = ali[1].split(' ');
        final scresu = await scrapeBClink('https://' + alib[0]);
        if (scresu.isNotEmpty) {
          //pull the image as attachement

          if (scresu.last is Uint8List && attachement.isEmpty) {
            attachement = scresu.last;
            //patch url title pair
            text = '[' + scresu[0] + '](' + scresu[1] + ') ';
          } else {
            text = scresu[2];
          }

          if (!bcHash && scresu.contains('bandcamp')) tagBc = true;
          bool sk = false;
          alib.forEach((element) {
            if (!sk) {
              sk = true;
              return;
            }
            text += ' ' + element;
          });

          if (!bcHash && tagBc) text += ' #bandcamp';
        }
      }
    } //dont touch markdown links

    if (attachement.isEmpty) {
      return await _io!
          .newThread(baseUrl, text, shortLinkOfPrevious,
              _latestCapsule!.katamari, _nick, _avatar)
          .then((value) {
        if (value.isEmpty) return null;
        return updateLatestCapsule(value);
      });
    }
    //detect attachmnet filetpye
    bool isAudio = false;
    bool isVideo = false;
    bool isJpeg = false;
    bool isPng = false;
    bool isGif = false;
    bool isImage = false;

    var mty = await lookupMimeType('', headerBytes: attachement);
    int atType = 0;
    if (mty == null) mty = attachmentType;

    switch (mty) {
      case 'image/jpeg':
        isJpeg = true;
        isImage = true;
        break;
      case 'image/png':
        isPng = true;
        isImage = true;
        break;
      case 'image/gif':
        isGif = true;
        isImage = true;

        break;
      case 'video/mp4':
        isVideo = true;
        break;
      case 'audio/mpeg':
        isAudio = true;
        break;
    }
    if (mty.startsWith('video')) {
      isVideo = true;
      atType = 1;
    }
    if (mty.startsWith('audio')) {
      isAudio = true;
      atType = 2;
    }

    if (!isImage && !isAudio && !isVideo) {
      return null;
    }

    String cid = '';
    if ((isJpeg || isPng) && attachement.lengthInBytes > 200000) {
      attachement = await compressImageIsolate(attachement.toList()).bake();
      print('compressed');
    }
    if (isVideo && attachement.lengthInBytes > (5 * 1024000)) {
      return null;
    }
    if (isAudio && attachement.lengthInBytes > (5 * 1024000)) {
      return null;
    }
    //duplicate, file inserted on different stage of ipfs queue
    if (isGif) {
      print('gif:: ' + attachement.lengthInBytes.toString());
    }
    //final attXX = await xxHash(attachement).bake();
    final attXX = _xxhc[attachement.lengthInBytes] != null
        ? _xxhc[attachement.lengthInBytes]!
        : await xxHash(attachement).bake();
    _xxhc[attachement.lengthInBytes] = attXX;

    cid = _ipfsQ.findCid(attXX);

    if (cid.isNotEmpty) {
      //text += ' !&' + cid + ' ';
      xxC.pushCidData(cid, attachement);

      try {
        final image = img.decodeImage(attachement);
        final blurHash = BlurHash.encode(image!, numCompX: 4, numCompY: 3);
        text += ' !|' + blurHash.hash + ' ';
        print('blurhashed');
      } catch (e) {}

      return await _io!
          .newThreadWithIPFSAttachement(
              baseUrl,
              text,
              shortLinkOfPrevious,
              _latestCapsule!.katamari,
              _nick,
              attachement.lengthInBytes,
              attXX,
              cid,
              _avatar,
              atType)
          .then((value) {
        if (value.isEmpty) return null;
        return updateLatestCapsule(value);
      });
    }

    //no local cid mapping found.

    print('attachemnt len ' + attachement.lengthInBytes.toString());
    String pushCid = await _ipfsQ.pushToPinataIPFS('', '', attachement);

    if (pushCid.isEmpty) return null;
    //text += ' !&' + pushCid + ' ';
    xxC.pushCidData(pushCid, attachement);
    try {
      final image = img.decodeImage(attachement);
      final blurHash = BlurHash.encode(image!, numCompX: 4, numCompY: 3);
      text += ' !|' + blurHash.hash + ' ';
    } catch (e) {}
    return _io!
        .newThreadWithIPFSAttachement(
            baseUrl,
            text,
            shortLinkOfPrevious,
            _latestCapsule != null ? _latestCapsule!.katamari : Uint8List(0),
            currentNick,
            attachement.lengthInBytes,
            attXX,
            pushCid,
            currentAvatar,
            atType)
        .then((value) {
      if (value.isEmpty) return null;
      return updateLatestCapsule(value);
    });
  }

  ModeratorCapsule? newGroup(String text, Uint8List attachement) {}
  ModeratorCapsule? newGroupChild(String text, Uint8List attachement) {}

  Future<List<ModeratorEntry>?> newReply(
      String text,
      String nick,
      String avatar,
      String rootShortlink,
      String ppShortlink,
      Uint8List attachement,
      String attachmentType) async {
    Uint8List nt;
    if (text.contains('https://')) {
      bool bcHash = text.contains('#bandcamp');
      bool tagBc = false;
      if (!text.startsWith('[')) {
        final ali = text.split('https://');
        final alib = ali[1].split(' ');
        final scresu = await scrapeBClink('https://' + alib[0]);
        if (scresu.isNotEmpty) {
          //pull the image as attachement
          if (scresu.last is Uint8List && attachement.isEmpty) {
            attachement = scresu.last;
            //patch url title pair
            text = '[' + scresu[0] + '](' + scresu[1] + ') ';
          } else {
            text = scresu[2];
          }

          if (!bcHash && scresu.contains('bandcamp')) tagBc = true;
          bool sk = false;
          alib.forEach((element) {
            if (!sk) {
              sk = true;
              return;
            }
            text += ' ' + element;
          });

          if (!bcHash && tagBc) text += ' #bandcamp';
        }
      }
    } //dont touch markdown links

    if (attachement.isEmpty) {
      nt = await _io!.newReply(
          baseUrl,
          text,
          rootShortlink,
          ppShortlink,
          _latestCapsule != null ? _latestCapsule!.katamari : Uint8List(0),
          nick,
          avatar);

      if (nt.isEmpty) return null;
      //return empty list with dup
      final u = updateLatestCapsule(nt);
      if (u != null && u.isEmpty) return [];

      _dirty = true;
      return threadByShortLink(rootShortlink).toList();
    }
    String cid = '';
    //detect attachmnet filetpye
    bool isAudio = false;
    bool isVideo = false;
    bool isJpeg = false;
    bool isPng = false;
    bool isGif = false;
    bool isImage = false;

    var mty = await lookupMimeType('', headerBytes: attachement);

    if (mty == null) {
      mty = attachmentType;
    }
    int atType = 0;
    switch (mty) {
      case 'image/jpeg':
        isJpeg = true;
        isImage = true;
        break;
      case 'image/png':
        isPng = true;
        isImage = true;
        break;
      case 'image/gif':
        isGif = true;
        isImage = true;

        break;
      case 'video/mp4':
        isVideo = true;
        atType = 1;
        break;
      case 'audio/mpeg':
        isAudio = true;
        atType = 2;
        break;
    }
    if (mty.startsWith('video')) {
      isVideo = true;
      atType = 1;
    }
    if (mty.startsWith('audio')) {
      isAudio = true;
      atType = 2;
    }
    if (!isImage && !isAudio && !isVideo) {
      return null;
    }

    if ((isJpeg || isPng) && attachement.lengthInBytes > 150000) {
      attachement = await compressImageIsolate(attachement.toList()).bake();
      print('compressed');
    }
    if (isVideo && attachement.lengthInBytes > (5 * 1024000)) {
      return null;
    }
    if (isAudio && attachement.lengthInBytes > (5 * 1024000)) {
      return null;
    }
    //duplicate, file inserted on different stage of ipfs queue
    if (isGif) {
      print('gif:: ' + attachement.lengthInBytes.toString());
    }

    //duplicate, file inserted on different stage of ipfs queue
    //final attXX = attachement.hashCode;
    final attXX = _xxhc[attachement.lengthInBytes] != null
        ? _xxhc[attachement.lengthInBytes]!
        : await xxHash(attachement).bake();
    _xxhc[attachement.lengthInBytes] = attXX;

    cid = _ipfsQ.findCid(attXX);

    if (cid.isNotEmpty) {
      //text += ' !&' + cid + ' ';
      xxC.pushCidData(cid, attachement);
      try {
        final image = img.decodeImage(attachement);
        final blurHash = BlurHash.encode(image!, numCompX: 4, numCompY: 3);
        text += ' !|' + blurHash.hash + ' ';
      } catch (e) {}

      return await _io!
          .newReplyWithIPFSAttachement(
              baseUrl,
              rootShortlink,
              text,
              ppShortlink,
              _latestCapsule!.katamari,
              nick,
              attachement.lengthInBytes,
              attXX,
              cid,
              nick,
              atType)
          .then((value) {
        if (value.isEmpty) return null;
        //gives empty list on dup
        return updateLatestCapsule(value);
      });
    }

    //no local cid mapping found.

    print('attachemnt len ' + attachement.lengthInBytes.toString());
    String pushCid = await _ipfsQ.pushToPinataIPFS('', '', attachement);

    if (pushCid.isEmpty) return null;
    //text += ' !&' + pushCid + ' ';
    xxC.pushCidData(pushCid, attachement);
    try {
      final image = img.decodeImage(attachement);
      final blurHash = BlurHash.encode(image!, numCompX: 4, numCompY: 3);
      text += ' !|' + blurHash.hash + ' ';
    } catch (e) {}
    return _io!
        .newReplyWithIPFSAttachement(
            baseUrl,
            rootShortlink,
            text,
            ppShortlink,
            _latestCapsule!.katamari,
            nick,
            attachement.lengthInBytes,
            attXX,
            pushCid,
            avatar,
            atType)
        .then((nt) {
      if (nt.isEmpty) return null;

      updateLatestCapsule(nt);
      _dirty = true;
      return threadByShortLink(rootShortlink).toList();
    });
  }

  Future<int> newReplyToQueue(
      String text,
      String shortLinkOfRoot,
      String shortLinkOfPP,
      Uint8List attachement,
      String attachmentType,
      String nick,
      String avatar) async {
    final ogm = _ogMessages.addReply(nick, avatar, text, shortLinkOfRoot,
        shortLinkOfPP, now, attachement, '', attachmentType);
    if (ogm == null) return -1;
    if (!_online) {
      //just push to disk when no net availablus
      _ogMessages.pushToDiskQueue(ogm.timestamp);
    }
    return ogm.timestamp;
  }

  Future<Uint8List?> readLocalFileByFilename(String filename) async {
    String cp = Directory.current.path;
    try {
      var did = await File(Directory.current.path + '/testImages/' + filename)
          .readAsBytes();
      return did;
    } catch (e) {
      return null;
    }
  }

  List<ModeratorEntry> insertToLocalEntrySet(ModeratorCapsule c) {
    List<ModeratorEntry> _nw = [];
    final es = ModeratorEntrySet.fromModeratorCapsule(c.threadEntries);

    if (es.all.isEmpty) return _nw;
    var i = 0;
    es.all.forEach((element) {
      if (_mutedAuthors.contains(element.author)) {
        return;
      }
      if (_deletedEntries.contains(element.keyHint)) {
        return;
      }
      //remember is seen
      print(element.hint);
      print(element.likes);
      element.flags.isSeen = _omni.isSeen(element.hint);

      if (data.add(element)) {
        _nw.add(element);
        i++;
      }
    });
    if (i == 0) return _nw; //no new entries

    Uint8List ul = ModeratorCapsuleHelper(data);
    if (ul.isNotEmpty) {
      _disk?.commitCapsule(ul);
    }
    return _nw;
  }

  LocalEntrySetPrime(Uint8List l) {
    final es = ModeratorEntrySet.fromModeratorCapsule(l);
    if (es.all.isEmpty) return;
    var i = 0;
    for (var ii in es.all) {
      if (_mutedAuthors.contains(ii.author)) continue;
      if (_deletedEntries.contains(ii.keyHint)) continue;
      if (data.add(ii)) {
        i++;
      }
    }
    if (i > 0) _dirty = true;
    print('LocalEntrySetPrime with ' + i.toString());
  }

  Uint8List ModeratorCapsuleHelper(ModeratorEntrySet mode) {
    List<ModeratorCapsuleItem> mcl = [];

    if (mode.all.isEmpty) {
      print('PANIC: moderator capsule helper, moderator entryset is empty');
      return Uint8List(0);
    }

    //pluggy backlinks

    //pluggy perpre

    //mangle

    for (var ee in mode.all) {
      mcl.add(ModeratorCapsuleItem(
          ee.flags.binary,
          0,
          ee.bl,
          ee.pp,
          ee.timeElapsedSincePostMs,
          ee.body.length,
          ee.bodyBytes,
          ee.shortLinkBytes.lengthInBytes,
          ee.shortLinkBytes,
          ee.keyHint,
          ee.avatar,
          ee.nick,
          ee.attachmentLink,
          0,
          ee.perceivedTimestamp,
          ee.likes,
          ee.shares,
          ee.views,
          ee.replycount));
    }

    var fppf = mcl.last;
    var fppfb = mcl.last.binary;
    ModeratorCapsuleItem popo = ModeratorCapsuleItem.fromBinary(fppfb);

    Uint8List enl = Uint8List(1064000);

    var wi = 0;
    for (var element in mcl) {
      if (wi < 3064000 / 0.8) {
        element.binary.forEach((element2) {
          enl[wi++] = element2;
        });
      }
    }

    return enl.sublist(0, wi);
  }

  List<List<dynamic>> _xxPullQueue = [];
  Set<int> _xxCheckedQueue = {};
  int _xxpqcou = 0;

  scanForCachedAttachements() async {
    //proto pull all attachemendos
    _xxpqcou = 0;

    if (_xxCheckedQueue.length < _xxPullQueue.length) return;
    if (data.all.isNotEmpty) {
      for (var element in data.all) {
        if (element.flags.attachements) {
          if (_xxCache.isXXcached(element.attachmentLinkAsXXint)) return;
          //print(element.shortLink);
          _xxPullQueue.add([element.attachmentLinkAsXXint, element.shortLink]);
        }
      }
      ;
    }
    if (_xxPullQueue.isEmpty) return;
  }

  primeCurrentViewWithCachedAttachments(ModeratorEntrySet m, int amount) {
    for (var ii = 0; ii < m.all.length; ii++) {
      if (!data.all[ii].flags.attachements) continue;
      if (_xxCache.isXXcached(m.all[ii].attachmentLinkAsXXint)) continue;
      _xxPullQueue.add([m.all[ii].attachmentLinkAsXXint, m.all[ii].shortLink]);
      amount--;
      if (amount < 1) break;
    }
  }

  pullForCachedAttachements() async {
    if (_xxPullQueue.isEmpty) return;

    Uint8List? xx = await _xxCache
        .pullXX(_xxPullQueue[_xxpqcou][0], _xxPullQueue[_xxpqcou][1], _server)
        .then((value) => null);
    _xxCheckedQueue.add(_xxpqcou);
    _xxpqcou++;
    if (xx == null) return;
    print(xx.lengthInBytes);
    _xxCheckedQueue.add(_xxpqcou);
    _xxpqcou++;
  }

  Future<Uint8List> satisfyXXwant(Uint8List _wants) async {
    if (_wants.isEmpty) return Uint8List(0);
    var s = 0;
    var b = new ByteData.view(_wants.buffer);
    var w = 0;
    while (w < _wants.lengthInBytes) {
      var xx = b.getUint64(w);
      if (_xxCache.isXXcached(xx)) {
        s = xx;
        break;
      }
      w += 8;
    }
    if (s == 0) return Uint8List(0);
    Uint8List? data = await _xxCache.pullXX(s, '', '');
    if (data != null) {
      //return _io!.satisfyXXwant(data);
      return data;
    }
    return Uint8List(0);
  }

  /*Future<String> pushToPinataIPFS(
      Uint8List data, pinata_api_key, pinata_secret_api_key) async {
    //add your pinata.cloud keys to PinataApiKeys.dart
    if (!haste) {
      return _ipfsQ.insertToQueue(xxHash(data), data);
    }
    return _io!.pushToPinataIPFS(pinata_api_key, pinata_secret_api_key, data);
  }*/
  Future<List<ModeratorEntry>?> dropEntry(String shortLink) async {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    if (target == null) return [];
    if (target.flags.moderate) {
      data.drop(target.keyHint);
      await _io?.newDrop(shortLink, _latestCapsule!.katamari);
      xxC.dropCID(target.attachmentLinkAsXXint, target.ipfsCid);
    }
    _dirty = true;
    return messages.all;
  }

  List<int> get droppedAttachments => data.droppedAttachements;
  int get attachmentsWeightInBytes => data.attachmentsWeightInBytes;

  Future<List<ModeratorEntry>?> dropEntryToQueue(String shortLink) async {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    if (target == null) return [];
    if (_deletedEntries.contains(target.keyHint)) return [];
    _deletedEntries.add(target.keyHint);
    //TODO kick off local storage too
    //xxC.delete(m.attachmentLink)
    //if it's your / your group, kill from serv too
    //await _disk?.commitPreferences(_io!.dartSessionId, _nick);
    await _disk?.commitDeletesPreferences(_deletedEntries.toList());
    if (target.flags.moderate) {
      _ogMessages.addDrop(target.shortLink, now);
    }
    //tell moderator to drop this

    _dirty = true;
    return messages.all;
  }

  bool _processingoutgoingMessageCacheSend = false;

  outgoingMessageCacheSendTimerRefresh() {
    if (_processingoutgoingMessageCacheSend) return null;
    if (!_online) return null;

    Timer(Duration(milliseconds: 312), () {
      outgoingMessageCacheSend;
    });
  }

  Future<List<ModeratorEntry>?> outgoingMessageCacheSend(int timestamp) async {
    if (_processingoutgoingMessageCacheSend) return null;
    //if (!_online) return null;
    //dont worry about online offline
    //if this request works, we are online
    var pi;
    if (timestamp != 0) {
      //we need a specific item
      pi = await _ogMessages.pickPendingItem(timestamp);
    } else {
      pi = _ogMessages.pendingItem;
    }

    if (pi == null) return null;
    _processingoutgoingMessageCacheSend = true;

    //if the request does not work, this prevents blocking the queue
    ModeratorEntry? target =
        findModeratorEntryByShortLink(pi.shortLinkOfPrevious);
    if (target == null) {
      _ogMessages.releasePendingItem(pi.timestamp);
      _processingoutgoingMessageCacheSend = false;
      return null;
    }
    _ogMessages.holdPendingItem(pi.timestamp);

    if (pi.type == 3) {
      final r = await dropEntry(target.shortLink);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        outgoingMessageCacheSendTimerRefresh();
        return r;
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }
    if (pi.type == 4) {
      final r = await muteEntry(target);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        outgoingMessageCacheSendTimerRefresh();
        return r;
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }
    if (pi.type == 5) {
      final r = await shareEntry(target);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        outgoingMessageCacheSendTimerRefresh();
        return r;
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }
    if (pi.type == 6) {
      final r = await likeEntry(target);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        outgoingMessageCacheSendTimerRefresh();
        return r;
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }
    if (pi.type == 7) {
      final r = await unlikeEntry(target);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        outgoingMessageCacheSendTimerRefresh();
        return r;
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }

    if (pi.type == 1) {
      final r = await newReply(
          pi.text,
          pi.nick,
          pi.avatar,
          pi.shortLinkOfPrevious,
          pi.shortLinkOfPP,
          pi.attachement,
          pi.attachmentType);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        _dirty = true;
        outgoingMessageCacheSendTimerRefresh();
        return r;
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }

    if (pi.type == 0 && pi.attachement.isEmpty) {
      final r = await newThread(baseUrl, pi.text, pi.attachement, pi.nick,
          pi.avatar, pi.shortLinkOfPrevious, pi.cid, pi.attachmentType);
      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        _dirty = true;
        outgoingMessageCacheSendTimerRefresh();
        return r;
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }

    if (pi.type == 0 && pi.attachement.isNotEmpty && pi.cid.isEmpty) {
      //inject pinata keys here
      final r = await newThread(baseUrl, pi.text, pi.attachement, pi.nick,
          pi.avatar, pi.shortLinkOfPrevious, pi.cid, pi.attachmentType);
      if (r != null) {
        //empty list means it's already sent
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        outgoingMessageCacheSendTimerRefresh();
        _dirty = true;
        return r;
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }
    _processingoutgoingMessageCacheSend = false;
  }

  Future<List<ModeratorEntry>?> muteEntry(ModeratorEntry target) async {
    Uint8List nt;
    nt = await _io!.newMute(target.shortLink, _latestCapsule!.katamari);

    if (nt.isEmpty) return null;
    return updateLatestCapsule(nt);
  }

  Future<int> muteAuthorToQueue(ModeratorEntry target) async {
    //if (m.flags.moderate) return false;
    if (_mutedAuthors.contains(target.keyHint)) return -1;
    _mutedAuthors.add(target.author);

    await _disk?.commitMutesPreferences(_mutedAuthors.toList());
    _dirty = true;
    final om = _ogMessages.addMute(target.shortLink, now);
    if (om == null) return -1;
    return om.timestamp;
  }

  Future<List<ModeratorEntry>?> likeEntry(ModeratorEntry target) async {
    Uint8List nt;
    //if (_omni.isLiked(target.hint)) return null;
    nt = await _io!.newLike(target.shortLink, _latestCapsule!.katamari);
    if (nt.isEmpty) return null;
    _omni.markLiked(target.hint);
    _dirty = true;
    return updateLatestCapsule(nt);
  }

  Future<int> likeEntryToQueue(ModeratorEntry target) async {
    Uint8List nt;
    if (_omni.isLiked(target.hint)) return -1;
    _omni.markLiked(target.hint);
    _dirty = true;
    final om = _ogMessages.addLike(target.shortLink, now);
    if (om == null) return -1;
    return om.timestamp;
  }

  Future<List<ModeratorEntry>?> unlikeEntry(ModeratorEntry target) async {
    Uint8List nt;
    if (_omni.isLiked(target.hint)) return null;
    nt = await _io!.newLike(target.shortLink, _latestCapsule!.katamari);
    if (nt.isEmpty) return null;
    _omni.unmarkLiked(target.hint);
    return updateLatestCapsule(nt);
  }

  Future<int> unlikeEntryToQueue(ModeratorEntry target) async {
    Uint8List nt;
    if (!_omni.isLiked(target.hint)) return -1;
    _omni.unmarkLiked(target.hint);
    _dirty = true;
    final om = _ogMessages.addunLike(target.shortLink, now);
    if (om == null) return -1;

    return om.timestamp;
  }

  Future<List<ModeratorEntry>?> shareEntry(ModeratorEntry target) async {
    Uint8List nt;
    nt = await _io!.newShare(target.shortLink, _latestCapsule!.katamari);
    _ogMessages.addShare(target.shortLink, now);
    if (nt.isEmpty) return null;
    return updateLatestCapsule(nt);
  }

  Future<int> shareEntryToQueue(ModeratorEntry target) async {
    final om = _ogMessages.addShare(target.shortLink, now);
    if (om == null) return -1;
    return om.timestamp;
  }

  ModeratorEntry? findModeratorEntryByShortLink(String shortLink) {
    return data.read(shortLink);
  }

  List<int> get stats {
    final s = [
      dataIn, //IOreceived
      dataOut, //IOsent
      dataPulledDisk, //disk pulled, total diskIO
      0, //memUsed
      0 //audience
    ];
    return s;
  }

  Future<List<ModeratorEntry>?> get login async {
    //TODO init stuff
    _io!.handshake(_server, _katamari).then((hs) async {
      if (hs.isNotEmpty) {
        dataIn += hs.lengthInBytes;
        //save prefs on sites layer not here
        //await _disk?.commitPreferences(
        //    _io!.dartSessionId, currentNick, currentAvatar);

        satisfyXXwant(_io!.xxWantsList);
        final haa = ModeratorCapsule.fromBinary(hs);
        //send the pending stuff if we got some
        refreshViaHTTP();
        return updateLatestCapsule(hs);
      }
    });
    return data.all;
  }

  Future<bool> get logout async {
    //TODO purge caches, send pending requests, etc

    return true;
  }
}

class compressImageIsolate {
  compressImageIsolate(this._attachementData);
  final List<int> _attachementData;
  Future<Uint8List> bake() async {
    final p = ReceivePort();
    // spawn the isolate and wait for it to complete
    await Isolate.spawn(_compressImageIsolate, p.sendPort);
    return await p.first;
  }

  Future<void> _compressImageIsolate(SendPort p) async {
    final dad = Uint8List.fromList(_attachementData);
    final input = ImageFile(
      rawBytes: dad,
      filePath: 'gaggi',
    );
    final output = compress(ImageFileConfiguration(
        input: input,
        config: Configuration(jpgQuality: 52, outputType: OutputType.jpg)));

    Isolate.exit(p, output.rawBytes);
  }
}

class xxHash {
  xxHash(this._xxData);
  final List<int> _xxData;

  Future<int> bake() async {
    final p = ReceivePort();
    // spawn the isolate and wait for it to complete
    await Isolate.spawn(_xxHashIsolate, p.sendPort);
    // get and return the result data
    return await p.first;
  }

  Future<void> _xxHashIsolate(SendPort p) async {
    //digest = _xxData.hashCode;
    final digest = xxHash64.convert(_xxData);
    final ii = digest.x.toInt();
    Isolate.exit(p, _xxData.hashCode);
  }
}
